/*
 * Copyright (c) 2009 HIS GmbH All Rights Reserved.
 *
 */
package de.his.core.util.openoffice;


import java.io.File;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.net.ConnectException;
import java.net.MalformedURLException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.common.collect.Lists;
import com.sun.star.beans.PropertyValue;
import com.sun.star.beans.PropertyVetoException;
import com.sun.star.beans.UnknownPropertyException;
import com.sun.star.beans.XPropertySet;
import com.sun.star.comp.helper.BootstrapException;
import com.sun.star.connection.ConnectionSetupException;
import com.sun.star.container.NoSuchElementException;
import com.sun.star.container.XNameAccess;
import com.sun.star.document.XDocumentInsertable;
import com.sun.star.frame.XComponentLoader;
import com.sun.star.frame.XStorable;
import com.sun.star.lang.IllegalArgumentException;
import com.sun.star.lang.IndexOutOfBoundsException;
import com.sun.star.lang.WrappedTargetException;
import com.sun.star.lang.XComponent;
import com.sun.star.style.BreakType;
import com.sun.star.style.XStyleFamiliesSupplier;
import com.sun.star.style.XStyleLoader;
import com.sun.star.table.XCell;
import com.sun.star.text.ControlCharacter;
import com.sun.star.text.XText;
import com.sun.star.text.XTextCursor;
import com.sun.star.text.XTextDocument;
import com.sun.star.text.XTextTable;
import com.sun.star.text.XTextTablesSupplier;
import com.sun.star.text.XTextViewCursor;
import com.sun.star.text.XTextViewCursorSupplier;
import com.sun.star.uno.UnoRuntime;
import com.sun.star.util.CloseVetoException;
import com.sun.star.util.XReplaceDescriptor;
import com.sun.star.util.XReplaceable;

import de.his.core.util.openoffice.OOoConnectionPool.OOoRuntimeNotFoundException;
import ooo.connector.server.OOoServer;

/**
 * Helper providing functionality of
 * - Offering a single network connection to OOo server
 * - Loading, temporarily storing and merging of OOo templates
 * - Merging pdf documents
 * - Processing placeholders within OOo documents
 * - Evaluating of conditional statements
 * - Replacements of german umlauts
 *
 * @author wahrendorff#his.de, lustig#his.de, reifschneider#his.de
 * @version $Revision: 1.10 $
 */

public class OOoHelper {
    /** Logger class */
    // protected static Logger logger = //logger.getLogger(OOoHelper.class);

    /** Buffer Size for zip files */
    private static final int BUFFER = 2048;

    /** Protokoll-String für Dateizugriff */
    public static final String PROTOCOL_FILE = "file:///";

    /** Protokoll-String für Open-Office Interna */
    public static final String PROTOCOL_PRIVATE = "private:";

    /** IF Clause Constants */
    private static final String IF_SEPARATOR = ",";

    private static final String IF_DEFAULTREPLACEMENT = "";

    private static final String IF_BRACKET_RIGHT = ")";

    private static final String IF_EQUAL = "==";

    private static final String IF_N_EQUAL = "!=";

    private static final String IF_QUOTE = "'";

    /** Is automatically determined if empty */
    private String installPath = "";

    //private OOoConnection conn = new OOoConnection();


    /*
     * Connection Ressources 
     */

    /** Server, if actively started from this bean */
    private OOoServer oooServer;

    /** Suffix / Prefix for Document Variables. */
    private final String VAR_PREFIX = "$$";// "«"; //default: "$$"

    private final String VAR_SUFFIX = "";// "»"; //default: ""

    /** Properties für den Dateizugriff */
    // private static PropertyValue[] HIDDEN_READ_ONLY_PROPERTY = new PropertyValue[] { makeProperty("ReadOnly", Boolean.TRUE), makeProperty("Hidden", Boolean.TRUE), makeProperty("UpdateDocMode", Short.valueOf(UpdateDocMode.NO_UPDATE)), makeProperty("MacroExecutionMode", Short.valueOf(MacroExecMode.NEVER_EXECUTE)) };
    private static PropertyValue[] HIDDEN_READ_ONLY_PROPERTY = new PropertyValue[] { makeProperty("ReadOnly", Boolean.TRUE), makeProperty("Hidden", Boolean.TRUE) };

    // private static PropertyValue[] HIDDEN_WRITE_PROPERTY = new PropertyValue[] { makeProperty("ReadOnly", Boolean.FALSE), makeProperty("Hidden", Boolean.TRUE), makeProperty("UpdateDocMode", Short.valueOf(UpdateDocMode.NO_UPDATE)), makeProperty("MacroExecutionMode", Short.valueOf(MacroExecMode.NEVER_EXECUTE)) };
    private static PropertyValue[] HIDDEN_WRITE_PROPERTY = new PropertyValue[] { makeProperty("ReadOnly", Boolean.FALSE), makeProperty("Hidden", Boolean.TRUE) };

    /** allowed content types (file formats) for Reports */
    private static final Map<String, String> contentTypes = setContentTypes();

    /** Anzahl der Millisekunden pro Tag */
    private static BigDecimal millisPerDay = BigDecimal.valueOf(1000 * 60 * 60 * 24);

    /** Differenz der Basis von OpenOffice-Date (30.12.1899) und Java-Date (January 1, 1970, 00:00:00 GMT) in Tagen */
    private static BigDecimal dateDifferenceDays = BigDecimal.valueOf(25569);

    /** Single Connection to be used by all OOo reports  */
    private OOoConnection oooConnection;

    /** Counter*/
    private int lastIndex;

    /*
     * Connection Properties
     */

    private String hostname = "localhost";//Conf.getString(Conf.CS.SYS.REPORTING_OO_ENDPOINT, "localhost");

    private Integer port = Integer.valueOf(8100);//Conf.getInteger(Conf.CS.SYS.REPORTING_OO_ENDPOINT_PORT, 8100);

    // count of connection retries before timeout, waits 1s (1000ms) between retries
    private Integer connectRetry = Integer.valueOf(60);//Conf.getInteger(Conf.CS.SYS.REPORTING_OO_ENDPOINT_CONNECT_RETRY, 60);


    /** Template File Path to load  */
    private String TEMPLATE_FILE_PATH;// = "TPLAlumni.odt";

    /** Template File to load  */
    //   private File TEMPLATE_FILE;

    /**
     * Copies the cell properties of a cell to another cell in the same table.
     *
     *
     * @param cellNameFrom name of the cell the properties are copied from
     * @param cellNameTo name of the cell the properties are copied to
     * @param xTable XTextTable
     */
    public static void copyCellProperties(String cellNameFrom, String cellNameTo, XTextTable xTable) {
        XText xCellTextFrom = UnoRuntime.queryInterface(XText.class, xTable.getCellByName(cellNameFrom));
        XTextCursor xTCursorFrom = xCellTextFrom.createTextCursor();
        XPropertySet xCursorPropsfrom = UnoRuntime.queryInterface(XPropertySet.class, xTCursorFrom);

        XText xCellTextTo = UnoRuntime.queryInterface(XText.class, xTable.getCellByName(cellNameTo));
        XTextCursor xTCursorTo = xCellTextTo.createTextCursor();
        XPropertySet xCursorPropsTo = UnoRuntime.queryInterface(XPropertySet.class, xTCursorTo);
        try {
            xCursorPropsTo.setPropertyValue("CharWeight", xCursorPropsfrom.getPropertyValue("CharWeight"));
        } catch (UnknownPropertyException e) {
            e.printStackTrace();
        } catch (PropertyVetoException e) {
            e.printStackTrace();
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
        } catch (WrappedTargetException e) {
            e.printStackTrace();
        }
    }


    /**
     * Finds a cell by its cell text in a text table and return the cells name.
     * Matches exactly. First match is returned.
     *
     * @param cellText cellText
     * @param textTable textTable
     * @return cellname
     */
    public static String findColumnNameByText(String cellText, XTextTable textTable) {
        String cellName = null;
        String cellNames[] = textTable.getCellNames();
        for (int i = 0; i < cellNames.length; i++) {
            XCell xCell = textTable.getCellByName(cellNames[i]);
            XText xTextCell = UnoRuntime.queryInterface(XText.class, xCell);
            String strText = xTextCell.getString();
            if (cellText.equals(strText)) {
                // cellname has min length 2
                return cellNames[i].substring(0, 1);
            }
        }
        return cellName;
    }

    /**
     * Searches the indexed table collection in XTextDocument and gets a text table by its name.
     * First match is returned.
     *
     * @param document XTextDocument
     * @param tableName Table name
     * @return XTextTable, null if not found
     * @throws WrappedTargetException
     * @throws IndexOutOfBoundsException
     */
    public static XTextTable findTextTable(XTextDocument document, String tableName) throws WrappedTargetException, IndexOutOfBoundsException {
        XTextTablesSupplier xTablesSupplier = UnoRuntime.queryInterface(XTextTablesSupplier.class, document);
        XNameAccess xNamedTables = xTablesSupplier.getTextTables();
        String[] xTableNames = xNamedTables.getElementNames();
        Object table = null;
        XTextTable xTable = null;
        for (String xTableName : xTableNames) {
            if (tableName.equals(xTableName)) {
                try {
                    table = xNamedTables.getByName(xTableName);
                    xTable = UnoRuntime.queryInterface(XTextTable.class, table);
                    return xTable;
                } catch (NoSuchElementException e) {
                    //logger.error(String.format("Could not get table with name %s in document.", xTableName));
                    return null;
                }
            }
        }
        return xTable;
    }

    /**
     * Used during Initialization
     * @return map from File-Formats to MimeTypes
     */
    private static final Map<String, String> setContentTypes() {
        Map<String, String> contentTypesRaw = new HashMap<String, String>();
        contentTypesRaw.put("odt", "application/vnd.oasis.opendocument.text");
        contentTypesRaw.put("pdf", "application/pdf");
        contentTypesRaw.put("rtf", "text/rtf");
        contentTypesRaw.put("doc", "application/msword");
        contentTypesRaw.put("html", "text/html");
        contentTypesRaw.put("zip", "application/zip");

        return contentTypesRaw;
    }


    /**
     * returns MimeType for formats: odt, pdf, rtf, doc and html
     * @param format
     * @return String MimeType
     */
    public String getContentType(String format) {
        String result = contentTypes.get(format);
        if (result != null) {
            return result;
        } else {
            return "application/octet-stream";
        }
    }


    /**
     * Propertys for handling oo docs hidden, means: without gui
     * @param name
     * @param value
     * @return
     */
    private static PropertyValue makeProperty(String name, Object value) {
        PropertyValue property = new PropertyValue();
        property.Name = name;
        property.Value = value;
        return property;
    }


    /**
     * Create a new, empty writer doc
     * @return new, empty writer doc as XTextDocument Interface or null
     * @throws ConnectException
     * @throws OOoRuntimeNotFoundException
     * @throws Exception
     */
    public XTextDocument newWriterDoc() throws ConnectException, OOoRuntimeNotFoundException, Exception {
        return loadTextDocument(PROTOCOL_PRIVATE, "factory/swriter", false);
    }


    /**
     * load a Writer document
     * @param pathToFile
     * @return writer doc as XTextDocument Interface or null
     * @throws ConnectException
     * @throws OOoRuntimeNotFoundException
     * @throws Exception
     */
    public XTextDocument loadWriterDoc(String pathToFile) throws ConnectException, OOoRuntimeNotFoundException, Exception {
        return loadTextDocument(PROTOCOL_FILE, pathToFile, false);
    }


    /**
     * load a writer doc in read only mode
     * @param pathToFile
     * @return readonly document for the Template set via 'setTemplate(...)'
     * @throws OOoRuntimeNotFoundException
     * @throws ConnectionSetupException
     * @throws BootstrapException
     * @throws IOException
     */
    public XTextDocument loadTemplateDoc(String pathToFile) throws OOoRuntimeNotFoundException, ConnectionSetupException, IOException, BootstrapException {
        return loadTextDocument(PROTOCOL_FILE, pathToFile, true);
    }



    /**
     * @return readonly document for the Template set via 'setTemplate(...)'
     * @throws ConnectException
     * @throws OOoRuntimeNotFoundException
     * @throws Exception
     */
    public XTextDocument loadTemplateDoc() throws ConnectException, OOoRuntimeNotFoundException, Exception {
        return loadTextDocument(PROTOCOL_FILE, getTemplatePath(), true);
    }

    /**
     * @param protocol
     * @param url
     * @param readonly
     * @return Text Document identified by given protocol and url. Optionally open readonly
     * @throws ConnectionSetupException
     * @throws BootstrapException
     * @throws IOException
     */
    private XTextDocument loadTextDocument(String protocol, String url, boolean readonly)
                    throws OOoRuntimeNotFoundException, ConnectionSetupException, IOException, BootstrapException {
        return loadTextDocument(getConnection().getComponentLoader(), protocol, url, readonly);
    }


    /**
     * @param loader
     * @param protocol
     * @param url
     * @param readonly
     * @return Text Document from XComponentLoader identified by given protocol and url. Optionally open readonly
     */
    public static XTextDocument loadTextDocument(XComponentLoader loader, String protocol, String url, boolean readonly) {
        Object document = null;
        XTextDocument xTextDoc = null;
        try {
            String ressource = protocol + url;
            //logger.info("loadTextDocument(): ressource: " + ressource);
            document = loader.loadComponentFromURL(ressource, "_blank", 8, readonly ? HIDDEN_READ_ONLY_PROPERTY : HIDDEN_WRITE_PROPERTY);
            xTextDoc = UnoRuntime.queryInterface(XTextDocument.class, document);
        } catch (IllegalArgumentException iae) {
            //logger.error("OO IA-ERROR: " + iae.getMessage());
            //logger.error("loadTextDocument(): url: " + url);
            return null;
        } catch (com.sun.star.io.IOException e) {
            //logger.error("OO IO-ERROR: " + e.getMessage());
            //logger.error("loadTextDocument(): url: " + url);
            return null;
        } catch (com.sun.star.lang.DisposedException e) {
            //logger.error("OO Bridge-ERROR: " + e.getMessage());
            //logger.error("loadTextDocument(): url: " + url);
            return null;
        }
        return xTextDoc;
    }

    /**
     * @param <T>
     * @param myClass
     * @param myObject
     * @return <?>
     */
    public static <T> T queryInterface(Class<T> myClass, Object myObject) {
        T xLoader = UnoRuntime.queryInterface(myClass, myObject);
        return xLoader;
    }


    /**
     * stores a temporary File
     * @param source
     * @return File for stored Document
     */
    public File storeTempFile(XTextDocument source) {
        return storeTempFile(source, "oo_", ".odt");
    }


    /**
     * This functions stores an XTextDocument in temporäry filesystem..
     * @param source
     * @param prefix
     * @param suffix
     * @return File for stored Document
     */
    public File storeTempFile(XTextDocument source, String prefix, String suffix) {
        XStorable storable = queryInterface(XStorable.class, source);
        File tempFile = null;
        PropertyValue[] aMediaDescriptor = new PropertyValue[2];
        aMediaDescriptor[0] = createPropertyValue("Overwrite", Boolean.valueOf(true));

        try {
            tempFile = File.createTempFile(prefix, suffix);
        } catch (IOException ioe) {
            //logger.error("Could not create Temp File!");
        }

        try {
            if (tempFile != null) storable.storeToURL("file:///" + tempFile.getAbsolutePath().replace("\\", "/"), aMediaDescriptor);
        } catch (com.sun.star.io.IOException e) {
            //
            //logger.error(e.getLocalizedMessage());
        }

        return tempFile;
    }

    /**
     * Returns an Array of PDFFilter Options (PropertyValues)
     * see http://wiki.openoffice.org/wiki/API/Tutorials/PDF_export
     * for more examples
     * @return PropertyValue[]
     */
    //@SuppressWarnings("unused")
    private static PropertyValue[] createPdfFilters(Long pdfVersion) {
        PropertyValue[] pdfFilterProperties = new PropertyValue[3];
        /**
         * Specifies the version of PDF to emit.
         *
         *  Possible values:
         *  
         *      0 = PDF 1.4 (default selection).
         *      1 = PDF/A-1 (ISO 19005-1:2005) 
         */

        pdfFilterProperties[0] = makeProperty("SelectPdfVersion", pdfVersion);

        /**
         * Determines if PDF are created by using special tags also known as Tagged PDF. 
         * - Für barrierefreie PDFs
         */


        pdfFilterProperties[1] = makeProperty("InitialView", new Integer(1));

        pdfFilterProperties[2] = makeProperty("UseTaggedPDF", new Boolean(true));



        return pdfFilterProperties;
    }


    /**
     * @return String
     */
    public String getTemplatePath() {
        return this.TEMPLATE_FILE_PATH;
    }

    /**
     *
     * @param path
     */
    public void setTemplatePath(String path) {
        this.TEMPLATE_FILE_PATH = path;
    }

    /**
     * closes an given OOo DocumentComponent
     * @param oDocToClose
     * @throws CloseVetoException
     */
    public void close(Object oDocToClose) throws CloseVetoException {
        com.sun.star.util.XCloseable xCloseable = UnoRuntime.queryInterface(com.sun.star.util.XCloseable.class, oDocToClose);

        if (xCloseable != null) {
            xCloseable.close(false);
        } else {
            XComponent xComp = UnoRuntime.queryInterface(com.sun.star.lang.XComponent.class, oDocToClose);
            if (xComp != null) {
                xComp.dispose();
            }
        }
        return;
    }

    /**
     * Converts a XTextDocument to given format and stores it as Temp File.
     * @param xTextDocument
     * @param prefix
     * @param suffix - defines the Document Outputformat (odt, doc, rtf, pdf, xhtml)
     * @param pdfVersion - defines the PDF Version 0 for 1.4 or 1 for A-1a
     * @return File
     */
    public static File storeAsTemp(XTextDocument xTextDocument, String prefix, String suffix, Long pdfVersion) {
        /**
         * create Export Filter
         */
        //logger.debug("convert and store to temp");
        PropertyValue[] aMediaDescriptor = new PropertyValue[2];
        try {
            aMediaDescriptor[0] = getFilterProperty(suffix);

            /**
             *  Filters for pdf generation,
             *  for more options see http://wiki.openoffice.org/wiki/API/Tutorials/PDF_export
             */
            if (suffix.toLowerCase().equals("pdf")) {
                // creates PDFFilters
                aMediaDescriptor[1] = makeProperty("FilterData", createPdfFilters(pdfVersion));
            }

        } catch (Exception e) {
            //logger.error(e.getLocalizedMessage());
            return null;
        }

        XStorable xStorable = queryInterface(XStorable.class, xTextDocument);

        File tempFile = null;
        try {
            tempFile = File.createTempFile(prefix, "." + suffix);
        } catch (IOException ioe) {
            System.out.println("Could not create Temp File '" + prefix + "." + suffix + "' ...");
            return null;
        }

        try {
            if (tempFile != null) xStorable.storeToURL("file:///" + winPathBugfix(tempFile.getAbsolutePath().replace("\\", "/")), aMediaDescriptor);
        } catch (com.sun.star.io.IOException e) {
            //logger.error("Could not store to temp File " + e.getLocalizedMessage());
            return null;
        }

        return tempFile;
    }

    /**
     * A workaround for a java bug in Windows.
     * normally java should begin a windows path with Driveletter and :
     * but it does not, it puts an / in front of the driveletter as if we are on a unix plattform.
     * TODO: Wirklich noch nötig, nachdem 'getServerRootPath' umgestellt wurde?
     *
     * This functions cuts the trailing / in Front of a path, if, and only
     * if we are on a Windows plattform
     * and the first Char is actually /
     * @param path - String
     * @return cleanPath - String
     */
    public static String winPathBugfix(String path) {
        String osname = System.getProperty("os.name");
        String cleanPath = path;
        if (osname.startsWith("Windows")) {
            if (path.startsWith("/")) cleanPath = path.substring(1, path.length());
        }
        return cleanPath;
    }


    /**
     * replaces all placeholders in the template with their values
     * @param source
     * @param keyValueMap
     * @return XTextDocument Interface
     */
    public XTextDocument replacePlaceholders(XTextDocument source, Map<String, Object> keyValueMap) {
        XReplaceable xReplaceableDoc = queryInterface(XReplaceable.class, source);
        XReplaceDescriptor xReplaceDescr = xReplaceableDoc.createReplaceDescriptor();
        SimpleDateFormat simpleDateFormat = new SimpleDateFormat("dd.MM.yyyy");
        keyValueMap.put("Datum", simpleDateFormat.format(new Date()));

        Set<String> keySet = keyValueMap.keySet();
        for (String key : keySet) {
            xReplaceDescr.setSearchString(VAR_PREFIX + key + VAR_SUFFIX);
            xReplaceDescr.setReplaceString((String) keyValueMap.get(key));
            try {
                xReplaceDescr.setPropertyValue("SearchWords", Boolean.TRUE);
            } catch (UnknownPropertyException e) {
                //logger.error("PropertyValue SearchWords unbekannt");
                //logger.error(e.getLocalizedMessage());
            } catch (PropertyVetoException e) {
                //logger.error(e.getLocalizedMessage());
            } catch (IllegalArgumentException e) {
                //logger.error(e.getLocalizedMessage());
            } catch (WrappedTargetException e) {
                //logger.error(e.getLocalizedMessage());
            }

            int xFound = xReplaceableDoc.replaceAll(xReplaceDescr);
            //logger.debug(String.format("Platzhalter: %s  ersetzt durch: %s %d mal", xReplaceDescr.getSearchString(), xReplaceDescr.getReplaceString(), Integer.valueOf(xFound)));
        }

        return source;
    }


    /**
     * merges a Set of pdf Files to one big pdf , erases the sourcedocuments and returns a reference to the big File
     * @param pdfSet Set<File>
     * @param reportContext String
     * @return File
     */



    /**
     * zips a list of Files
     * @param fileList
     * @param jobName
     * @return File
     */


    /**
     * merges odt Files to one, it is likely to lose functions and complex layout.
     * @param fileSet
     * @param reportFormat
     * @param reportContext
     * @return File
     * @throws OOoConnectionException
     * @throws OOoMisconfigurationException
     * @throws OOoCommandException
     */
    public File mergeOdtSet(List<File> fileSet, String reportFormat, String reportContext) throws Exception {
        final XTextDocument tempXdoc = newWriterDoc();

        int i = 0;
        for (final File file : fileSet) {
            i++;
            final XTextViewCursorSupplier xViewCursorSupplier_sourceDoc = UnoRuntime.queryInterface(XTextViewCursorSupplier.class, tempXdoc.getCurrentController());

            final XTextViewCursor xTextViewCursor_sourceDoc = xViewCursorSupplier_sourceDoc.getViewCursor();
            xTextViewCursor_sourceDoc.gotoEnd(false);
            final XText xText = tempXdoc.getText();
            final XTextCursor textCursor = xText.createTextCursor();

            textCursor.gotoEnd(false);
            final XDocumentInsertable xDocInsertable = queryInterface(XDocumentInsertable.class, textCursor);
            final XStyleFamiliesSupplier xStyleFamiliesSupplier = UnoRuntime.queryInterface(XStyleFamiliesSupplier.class, tempXdoc);
            final XStyleLoader xStyleLoader = UnoRuntime.queryInterface(XStyleLoader.class, xStyleFamiliesSupplier.getStyleFamilies());
            final com.sun.star.beans.PropertyValue[] styleOptions = xStyleLoader.getStyleLoaderOptions();

            try {
                copyStyles(tempXdoc, file.getAbsolutePath());
                xDocInsertable.insertDocumentFromURL(PROTOCOL_FILE + winPathBugfix(file.getAbsolutePath()), styleOptions);
            } catch (com.sun.star.io.IOException | IllegalArgumentException e) {
                throw new Exception("OO ERROR: mergingOdtSet (copyStyles, insertDocumentFromURL)", e);
            }

            // only insert Pagebreak if another page is coming
            if (i < fileSet.size()) {
                insertPageBreak(tempXdoc);
            }
        }
        //        deleteFileSet(fileSet);
        return storeAsTemp(tempXdoc, reportContext.substring(0, 3), reportFormat, null);
    }

    /**
    * copies styles from an .odt document in the given path to the given XTextDocument Interface
    * @param targetDoc XTextDocument
    * @param sourcePath String (Linux like: '/home/tim/template.odt', Win like: 'D:/bill/template.odt'
    * @throws com.sun.star.io.IOException
    */
    private static void copyStyles(XTextDocument targetDoc, String sourcePath) throws com.sun.star.io.IOException {
        final XStyleFamiliesSupplier xStyleFamiliesSupplier = UnoRuntime.queryInterface(XStyleFamiliesSupplier.class, targetDoc);
        final XStyleLoader xStyleLoader = UnoRuntime.queryInterface(XStyleLoader.class, xStyleFamiliesSupplier.getStyleFamilies());
        final com.sun.star.beans.PropertyValue[] styleOptions = xStyleLoader.getStyleLoaderOptions();

        final String sTemplateUrl = PROTOCOL_FILE + winPathBugfix(sourcePath);

        // import styles from the template, Default behaviour is: load all, overwrite all, see API Docs
        xStyleLoader.loadStylesFromURL(sTemplateUrl, styleOptions);
    }

    /**
     * inserts a PageBreak in given XTextDocument Interfaces Document
     * @param xTextDoc
     * @throws OOoCommandException
     */
    private static void insertPageBreak(XTextDocument xTextDoc) throws Exception {
        // get TextCursorSupplier
        final XTextViewCursorSupplier xViewCursorSupplier_sourceDoc = UnoRuntime.queryInterface(XTextViewCursorSupplier.class, xTextDoc.getCurrentController());
        // get View Curser
        final XTextViewCursor xTextViewCursor_sourceDoc = xViewCursorSupplier_sourceDoc.getViewCursor();
        // jump to end of View
        xTextViewCursor_sourceDoc.gotoEnd(false);
        // get XText Interface from Document
        final XText xText = xTextDoc.getText();
        // create a TextCursor
        final XTextCursor textCursor = xText.createTextCursor();
        // jump to the End of all text
        textCursor.gotoEnd(false);
        // Get cursor property set
        final XPropertySet xPagebreakCursorProps = UnoRuntime.queryInterface(XPropertySet.class, textCursor);

        // Insert page break
        try {
            xPagebreakCursorProps.setPropertyValue("BreakType", BreakType.PAGE_AFTER);
        } catch (PropertyVetoException | IllegalArgumentException | WrappedTargetException | UnknownPropertyException e) {
            throw new Exception("OO Error: property access: BreakType", e);
        }

        // Insert paragraph break, needed for other elements than text, to keep their position
        try {
            xText.insertControlCharacter(textCursor, ControlCharacter.PARAGRAPH_BREAK, false);
        } catch (final IllegalArgumentException e) {
            throw new Exception("OO Error: insertControlCharacter", e);
        }
    }
    /**
    * copies styles from an .odt document in the given path to the given XTextDocument Interface
    * @param targetDoc XTextDocument
    * @param sourcePath String (Linux like: '/home/tim/template.odt', Win like: 'D:/bill/template.odt'
    */
    //    private void copyStyles(XTextDocument targetDoc, String sourcePath) {
    //        XStyleFamiliesSupplier xStyleFamiliesSupplier = UnoRuntime.queryInterface(XStyleFamiliesSupplier.class, targetDoc);
    //        XStyleLoader xStyleLoader = UnoRuntime.queryInterface(XStyleLoader.class, xStyleFamiliesSupplier.getStyleFamilies());
    //        com.sun.star.beans.PropertyValue[] styleOptions = xStyleLoader.getStyleLoaderOptions();
    //
    //        String sTemplateUrl = PROTOCOL_FILE + winPathBugfix(sourcePath);
    //
    //        // import styles from the template, Default behaviour is: load all, overwrite all, see API Docs
    //        try {
    //            xStyleLoader.loadStylesFromURL(sTemplateUrl, styleOptions);
    //        } catch (Exception e) {
    //            //logger.error(e.getLocalizedMessage());
    //        }
    //    }


    /**
     * inserts a PageBreak in given XTextDocument Interfaces Document
     * @param xTextDoc
     */
    //    private void insertPageBreak(XTextDocument xTextDoc) {
    //        // get TextCursorSupplier
    //        XTextViewCursorSupplier xViewCursorSupplier_sourceDoc = UnoRuntime.queryInterface(XTextViewCursorSupplier.class, xTextDoc.getCurrentController());
    //        // get View Curser
    //        XTextViewCursor xTextViewCursor_sourceDoc = xViewCursorSupplier_sourceDoc.getViewCursor();
    //        // jump to end of View
    //        xTextViewCursor_sourceDoc.gotoEnd(false);
    //        // get XText Interface from Document
    //        XText xText = xTextDoc.getText();
    //        // create a TextCursor
    //        XTextCursor textCursor = xText.createTextCursor();
    //        // jump to the End of all text
    //        textCursor.gotoEnd(false);
    //        // Get cursor property set
    //        XPropertySet xPagebreakCursorProps = UnoRuntime.queryInterface(XPropertySet.class, textCursor);
    //
    //        // Insert page break
    //        try {
    //            xPagebreakCursorProps.setPropertyValue("BreakType", BreakType.PAGE_AFTER);
    //        } catch (UnknownPropertyException e) {
    //            //logger.error(e.getLocalizedMessage());
    //        } catch (PropertyVetoException e) {
    //            //logger.error(e.getLocalizedMessage());
    //        } catch (IllegalArgumentException e) {
    //            //logger.error(e.getLocalizedMessage());
    //        } catch (WrappedTargetException e) {
    //            //logger.error(e.getLocalizedMessage());
    //        }
    //
    //        // Insert paragraph break, needed for other elements than text, to keep their position
    //        try {
    //            xText.insertControlCharacter(textCursor, ControlCharacter.PARAGRAPH_BREAK, false);
    //        } catch (IllegalArgumentException e) {
    //            //logger.error(e.getLocalizedMessage());
    //        }
    //    }


    /**
     * Refresh all Fields in the given Document
     * @param document
     */
    //    public void refreshAllFieds(XTextDocument document) {
    //        XTextFieldsSupplier xTextFieldsSupplier = UnoRuntime.queryInterface(XTextFieldsSupplier.class, document);
    //        // Access the TextFields and the TextFieldMasters collections
    //        XEnumerationAccess xEnumeratedFields = xTextFieldsSupplier.getTextFields();
    //        // afterwards we must refresh the textfields collection
    //        XRefreshable xRefreshable = UnoRuntime.queryInterface(XRefreshable.class, xEnumeratedFields);
    //        xRefreshable.refresh();
    //    }


    /**
     * replace some "evil" chars, a zipfile made with java cannot handle
     * zip files with utf-8 entrys will come with java7
     * @param filename
     * @return
     */
    //    private String germanUmlautsToAscii(String filename) {
    //        // //logger.debug("Umlaute vorher: " + filename);
    //        String newFilename = filename.replaceAll("Ü", "Ue").replaceAll("Ö", "Oe").replaceAll("Ä", "Ae").replaceAll("Ä", "Ae").replaceAll("ü", "ue").replaceAll("ö", "oe").replaceAll("ä", "ae").replaceAll("ß", "sz");
    //        // //logger.debug("Umlaute nacher: " + newFilename);
    //        return newFilename;
    //    }


    /**
     * deletes a Set of Files
     * @param fileSet
     */
    //    public void deleteFileSet(List<File> fileSet) {
    //        for (File file : fileSet) {
    //            if (file != null) {
    //                file.delete();
    //            }
    //        }
    //    }


    /**
     * used by OOoReportsTest and readSearchConfig
     * computes the path to WEB-INF Folder by reusing QIS functionality ;)
     * @return String
     */
    //    public static String getServerRootPath() {
    //        GlobalConfiguration.init();
    //        return GlobalConfiguration.getRPropDispatch().getProperty("QISSERVERURL");
    //    }



    /**
     * Returns a Filter PropertyValue for OpenOffice Export function.
     * @param format - extension for the file
     * @return Property Value
     * @throws Exception
     */
    public static PropertyValue getFilterProperty(String format) throws Exception {
        PropertyValue aFilter = new PropertyValue();
        aFilter = new PropertyValue();
        aFilter.Name = "FilterName";

        // writer Filter
        if (format.equals("odt"))
            aFilter.Value = "writer8";

        // pdf Filter
        else if (format.equals("pdf"))
            aFilter.Value = "writer_pdf_Export";

        // word Filter
        else if (format.equals("doc"))
            aFilter.Value = "MS Word 97";

        // Rich Text Format
        else if (format.equals("rtf"))
            aFilter.Value = "Rich Text Format";

        // HTML
        else if (format.equals("html"))
            aFilter.Value = "HTML (StarWriter)";
        else if (format.equals("xhtml"))
            aFilter.Value = "XHTML Writer File";
        else
            throw new Exception("Non supported Output Format");

        return aFilter;
    }


    /**
     * Returns a PropertyValue with given key and Value for OpenOffice calls, which need a PropertySet
     * @param key
     * @param value
     * @return PropertyValue
     */
    public PropertyValue createPropertyValue(String key, Object value) {
        PropertyValue aFilter = new PropertyValue();
        aFilter = new PropertyValue();
        aFilter.Name = key;
        aFilter.Value = value;
        return aFilter;
    }


    /**
     * evaluates if Conditions like: if('$$gender'=='w'){Sehr geehrte Frau}else{Sehr geehrter Herr};
     * @param source XTextDocument
     * @return XTextDocument
     */
    //    public XTextDocument evaluateIfConditions(XTextDocument source) {
    //        // Java Form Regex
    //        // String ifRegexJava = "\\$(if|IF|If|iF){1}\\(.*?\\)\\{(\\s*?.*?){0,}\\}\\$";
    //        // NK Form Regex
    //        String ifRegexNK = "\\$(if|IF|If|iF){1}\\(.*?,.*?(,.*?){0,1}\\)";
    //
    //        // used Regex
    //        String searchRegex = ifRegexNK;
    //        //logger.debug("IF-Clause Regex: " + searchRegex);
    //
    //        String replacement = IF_DEFAULTREPLACEMENT;
    //        String ifClause = null;
    //        String nestedIf = null;
    //
    //        String fullText = source.getText().getString();
    //        Pattern valuePattern = Pattern.compile(searchRegex);
    //        // valuePattern.CASE_INSENSITIVE
    //        Matcher valueMatcher = valuePattern.matcher(fullText);
    //
    //        while (valueMatcher.find() == true) {
    //
    //            ifClause = valueMatcher.group();
    //            // ifClause = ifClause.replace("\n", "").replace("\r", "");
    //            //logger.debug("IF-Clause match: " + ifClause);
    //            nestedIf = getNestedIf(ifClause);
    //            //logger.debug("Nested match: " + nestedIf);
    //            replacement = getReplacementForIfClause(nestedIf);
    //            //logger.debug("Replace " + nestedIf + " with: " + replacement);
    //
    //            // make the source Replaceable
    //            XReplaceable xReplaceableDoc = queryInterface(XReplaceable.class, source);
    //            XReplaceDescriptor xReplaceDescr = xReplaceableDoc.createReplaceDescriptor();
    //            try {
    //                xReplaceDescr.setPropertyValue("SearchCaseSensitive", Boolean.FALSE);
    //                // xReplaceDescr.setPropertyValue("SearchRegularExpression", Boolean.TRUE);
    //                // xReplaceDescr.setPropertyValue("SearchSimilarity", Boolean.TRUE);
    //                // xReplaceDescr.setPropertyValue("SearchSimilarityAdd", Short.valueOf("2"));
    //            } catch (Exception e) {
    //                // Auto-generated catch block
    //                //logger.error(e.getLocalizedMessage());
    //            }
    //
    //            // replace the if
    //            xReplaceDescr.setSearchString(nestedIf);
    //            xReplaceDescr.setReplaceString(replacement);
    //            //logger.debug("Replace String: " + xReplaceDescr.getReplaceString());
    //            xReplaceableDoc.replaceAll(xReplaceDescr);
    //
    //            if (!nestedIf.equals(ifClause)) {
    //                valueMatcher.reset();
    //                //logger.debug("RESET matcher: " + ifClause);
    //            }
    //
    //            fullText = source.getText().getString();
    //            valueMatcher = valuePattern.matcher(fullText);
    //        }
    //
    //        return source;
    //    }


    /**
     * gives back the deepest inner "nested if" or "original if", if no "nested if" was found
     * @param ifClause
     * @return String
     */
    //    private String getNestedIf(String ifClause) {
    //        if (lastIndex != 0 && lastIndex == ifClause.toLowerCase().lastIndexOf("$if")) {
    //            //throw new Exception("cs.sys.reporting.malformed_if_condition_in_document");
    //            //logger.error("cs.sys.reporting.malformed_if_condition_in_document...  last If Index: " + lastIndex);
    //        }
    //        lastIndex = ifClause.toLowerCase().lastIndexOf("$if");
    //        //logger.debug("last If Index: " + lastIndex);
    //        String innerIf = ifClause.substring(ifClause.toLowerCase().lastIndexOf("$if"), ifClause.length());
    //        return innerIf;
    //    }
    //
    //    /**
    //     * Gets replacement for non wrapped ifClause
    //     * @param ifClause
    //     * @return String
    //     */
    //    private String getReplacementForIfClause(String ifClause) {
    //        String[] ifValues = ifClause.split(IF_SEPARATOR);
    //
    //        String replacement = IF_DEFAULTREPLACEMENT;
    //        String ifCondition = ifValues[0].substring(4, ifValues[0].length());
    //        String ifReplacement = ifValues[1].trim().replace(IF_QUOTE, IF_DEFAULTREPLACEMENT);
    //        String elseReplacement = IF_DEFAULTREPLACEMENT;
    //
    //        if (ifValues.length > 2) {
    //            elseReplacement = ifValues[2].substring(0, ifValues[2].length() - 1).trim().replace(IF_QUOTE, IF_DEFAULTREPLACEMENT);
    //        } else {
    //            if (ifReplacement.endsWith(IF_BRACKET_RIGHT)) ifReplacement = ifValues[1].substring(0, ifValues[1].length() - 1);
    //        }
    //
    //        boolean equal = true;
    //        String[] conditionValues = ifCondition.split(IF_EQUAL);
    //
    //        if (conditionValues.length < 2) {
    //            equal = false;
    //            conditionValues = ifCondition.split(IF_N_EQUAL);
    //        }
    //        if (conditionValues.length < 2) {
    //            //logger.error("cs.sys.reporting.wrong_if_condition_in_document");
    //            // throw new ServiceException("cs.sys.reporting.wrong_if_condition_in_document");
    //        }
    //
    //        //logger.debug("IF-Condition is: " + ifCondition);
    //
    //        if ((equal == true && conditionValues[0].trim().equals(conditionValues[1].trim())) || (equal == false && !conditionValues[0].trim().equals(conditionValues[1].trim()))) {
    //            //logger.debug("IF-Condition " + ifCondition + " matches!");
    //            replacement = ifReplacement;
    //
    //        } else {
    //            //logger.debug("IF-Condition " + ifCondition + " does NOT match! Else replaces Condition");
    //            replacement = elseReplacement;
    //        }
    //        return replacement;
    //    }


    /**
     * @param date
     * @return Date formatted for OpenOffice (Days since 30.12.1899 with Time as Fractions of Days) as String with Comma
     */
    @SuppressWarnings("deprecation")
    public static String formatAsOpenOfficeDate(Date date) {
        BigDecimal millis = BigDecimal.valueOf(date.getTime() - date.getTimezoneOffset() * 60 * 1000); // Zeitzone und Winterzeit gegenüber Date(0) ausgleichen
        BigDecimal days = millis.divide(millisPerDay, 6, RoundingMode.UP); // 6 Nachkommastellen reichen für Sekunden-Genauigkeit aus
        days = days.add(dateDifferenceDays);

        return days.toPlainString().replace('.', ',');
    }


    /**
     * @return Connection to OpenOffice-Server
     * @throws ConnectionSetupException
     * @throws OOoRuntimeNotFoundException
     * @throws BootstrapException
     * @throws IOException
     * @throws ConnectException
     */
    public OOoConnection getConnection() throws OOoRuntimeNotFoundException, ConnectionSetupException, IOException, BootstrapException {
        // TODO: Try to fetch connection from a Pool
        OOoConnection conn = new OOoConnection();
        try {
            conn.connect(getConnectionString());
        } catch (ConnectException cex) {
            ////logger.warn(cex);
            try {
                // If there is no running server, try to start one and connect to it            
                tryToStartServer();
                //wait(1000);
                // wait until office is started
                for (int i = 0;; ++i) {
                    try {
                        conn.connect(getConnectionString());
                        break;
                    } catch (ConnectException ex) {
                        // Wait 500 ms, then try to connect again, but do not wait longer than 5 min (= 600 * 500 ms) total:
                        // 5 minutes is too long, make it customizable!
                        if (i == connectRetry.intValue()) {
                            throw ex;
                        }
                        try {
                            Thread.sleep(1000);
                        } catch (InterruptedException e) {
                            //logger.error(e);
                        }
                    }
                }
            } catch (OOoRuntimeNotFoundException ex) {
                throw ex;
            }
        }

        return conn;
    }

    /**
     * @throws IOException 
     * @throws BootstrapException 
     * 
     */
    private void tryToStartServer() throws OOoRuntimeNotFoundException, IOException, BootstrapException {

        // TODO: IllegalStateException, wenn oooServer-Variable nicht null ?

        if (!hostname.equals("localhost") && !hostname.equals("127.0.0.1")) {
            //throw exception if System tries to start OpenOffice on a remote Server, it is not possible. 
            //Remote Server HAVE to be started in advance
            //throw new de.his.appserver.service.ServiceException("cs.sys.ooo.server.cannotStartRemoteServer");
            //logger.error("OO ERROR: Cannot start remoteserver. Only localhost is possible atm.");
        }

        // Install-Path ermitteln, falls nicht gesetzt
        if (installPath == null || installPath.trim().isEmpty()) {
            installPath = InstallationFinder.getPath();
            if (installPath == null) {
                //logger.error("OO ERROR: Path to OO Executable could not be found.");
                throw new OOoRuntimeNotFoundException();
            }
        }

        // Versuch, Server zu starten
        try {
            // Create OOo server with additional -nofirststartwizard option
            List<String> oooOptions = Lists.newArrayList();//OOoServer.getDefaultOOoOptions();
            if (installPath.contains("libreoffice")) {
                // new options for LibreOffice
                oooOptions.add("--nologo");
                oooOptions.add("--nodefault");
                oooOptions.add("--norestore");
                oooOptions.add("--nocrashreport");
                oooOptions.add("--nolockcheck");
                oooOptions.add("--nofirststartwizard");
                oooOptions.add("--headless");
            } else {
                // compatibility options for OpenOffice
                oooOptions.add("-nologo");
                oooOptions.add("-nodefault");
                oooOptions.add("-norestore");
                oooOptions.add("-nocrashreport");
                oooOptions.add("-nolockcheck");
                oooOptions.add("-nofirststartwizard");
                oooOptions.add("-headless");
            }
            //logger.debug(""+oooOptions);
            oooServer = new OOoServer(installPath, oooOptions);
            oooServer.start(getAcceptOption());


        } catch (MalformedURLException ex) {
            ex.printStackTrace();
            throw ex;
        } catch (BootstrapException ex) {
            ex.printStackTrace();
            throw ex;
        } catch (IOException ex) {
            ex.printStackTrace();
            throw ex;
        }
    }


    /**
     * @return AcceptOptions for Server Start
     */
    protected String getAcceptOption() {
        String acceptOption;
        if (installPath.contains("libreoffice")) {
            acceptOption = "--accept=" + getConnectionString() + ";urp;StarOffice.ServiceManager;";
        } else {
            acceptOption = "-accept=" + getConnectionString() + ";urp;StarOffice.ServiceManager;";
        }
        return acceptOption;
    }


    /**
     * @return ConnectionString for Server Connection
     */
    protected String getConnectionString() {
        return "socket,host=" + hostname + ",port=" + port.toString() + ",tcpNoDelay=1";
    }

    //    @SuppressWarnings("unused")
    //    private void killServer() {
    //        OOoConnectionPool cp = getConnectionPool();
    //        close();
    //        cp.destroy();
    //        //logger.debug("OOoServer killed!");
    //
    //    }

    /**
     * Close this helper instance
     */
    public void close() {
        if (oooConnection != null) {
            oooConnection.disconnect();
            oooConnection = null;
        }
    }

}
